/* -----------------------------------------------------------------------------
 K03-21 Prevadec kodovani ceske diakritiky

 Martin Sloup, A04372
 FAV, INIB-INF
 msloup@students.zcu.cz
 arcao@arcao.com
 Soubor load_table.c
----------------------------------------------------------------------------- */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "function.h"
#include "structs.h"
#include "load_table.h"

CHARSTRUCT *table; /* Prekodovavaci tabulka */
int table_length; /* Velikost [bufferu] prekodovavaci tabulky */
int table_current_position; /* Aktualni pozice v prekodovavaci tabulce */ 
int loadfrom; /* Od jakeho ascii kodu tabulky budemem nacitat? */
int use_internal_encoding_from; /* Pouziti vestavene fce na prekodovani? 0=ascii, 1=utf8, 2=utf16 */
int use_internal_encoding_to; /* Pouziti vestavene fce na prekodovani? 0=ascii, 1=utf8, 2=utf16 */

/**
 * Tato funkce se stara o nacitani konverznich tabulek
 */
void load_table(char *filename, char mode) {
	FILE *fp;
	char sline[LINE_LENGTH];
	char *sline_work;
	char *sline_small;
	char **items;
	char *path;
	char *chyba;
	char *tmp;
	char *tmp2;

	int i = 0;
	int t_number = 0;
	int cont;

	/* prevedeme mode na velke pismeno */
	mode = toupper(mode);

	/**
	 * pokud jsme otevreli tabulku v modu Unicode a uz je neco v tabulce,
	 * pouzijeme specialni mod prirazeni
	 */
	if (mode == 'U' && table_length > 0) mode = 'S';
	
	/* opravime oddelovace adresaru */
	my_strtr(filename, "\\", PATH_SEP);
	my_strtr(filename, "/", PATH_SEP);
	
	/* otevreme tabulku */
	if ((fp = fopen(filename, "r")) == NULL) {
		error_handler(ENOENT, "load_table::fopen");
		fprintf(stderr, "Soubor '%s' se nepodarilo otevrit.\nSystem hlasi: %s.\n", filename, strerror(errno));
		exit(1);
	}
	
	while(fgets(sline, LINE_LENGTH, fp) != NULL) {
		cont = 0;
		
		/* odstranime mezery novou radku a tabulator ze zacatku a konce retezce */
		sline_work = trim((char *)sline);

		/* vytvorime promenou, a do ni vlozime radek konvertovany na ma pismena */
		new_string(sline_small, (strlen(sline_work) + 1));
		if (sline_small == NULL) continue;
		strcpy(sline_small, sline_work);
		lcase(sline_small);

		/* preskocime komentare */
		if (sline_small[0]=='#' && strstr(sline_small, "#alias ") == NULL && strstr(sline_small, "#include ") == NULL && strstr(sline_small, "#loadfrom ") == NULL && strstr(sline_small, "#internal ") == NULL) {
			memfree(sline_small);
			memfree(sline_work);
			continue;
		}
		/* a i prazdne radky */
		if (strlen(sline_small) == 0) {
			memfree(sline_small);
			memfree(sline_work);
			continue;
		}
		/* nastaveni nacitani znaku az od urcite ascii hodnoty */
		if (strstr(sline_small, "#loadfrom ") != NULL) {
			if (table_length == 0 || mode == 'S') {
				tmp = my_substr(sline_work, 10, 0);
				loadfrom = strtol(tmp, &chyba, 16);
				memfree(tmp);
			} else {
				/* pokud jiz mame neco nactene v tabulce, tak uz nemuzeme loadfrom pouzit */
				fprintf(stderr, "Varovani: #loadfrom jiz nelze pouzit, pokud je castecne nactena tabulka!\nSoubor: '%s'.\n", filename);
			}
			memfree(sline_small);
			memfree(sline_work);
			continue;
		}
		/* nastaveni kodovani pro interni funkce utf-8, utf-16 */
		if (strstr(sline_small, "#internal ") != NULL) {
			tmp = my_substr(sline_small, 10, 0);
			if (strcmp(tmp, "utf8") == 0) {
				if (mode == 'A') use_internal_encoding_from = 1;
				else use_internal_encoding_to = 1;
			} else if (strcmp(tmp, "utf16") == 0) {
				if (mode == 'A') use_internal_encoding_from = 2;
				else use_internal_encoding_to = 2;
			} else {
				fprintf(stderr, "Chyba v tabulce '%s'.\nV #internal bylo pouzito nezname kodovani! Program podporuje jen utf8 a utf16.", filename);
			}
			memfree(tmp);
			memfree(sline_small);
			memfree(sline_work);
			continue;
		}
		/* zpracovani vlozeneho souboru */
		if (strstr(sline_small, "#include ") != NULL) {
			tmp2 = my_substr(filename, 0, my_strrchr(filename, (char) ((PATH_SEP)[0])));
			tmp = my_substr(sline_work, 10, -1);
			new_string(path, (strlen(tmp2) + strlen(tmp) + 1));
			strcpy(path, tmp2);
			strcat(path, tmp);

			memfree(tmp);
			memfree(tmp2);
			memfree(sline_small);
			memfree(sline_work);
			
			/* rekurzivni volani funkce load_table */
			load_table(path, mode);
			memfree(path);
			continue;
		}
		if (cont==0) {
			/* nyni nasleduje vlastni zpracovani radku v tabulkovem souboru */
			/* vytahneme si ascii hodnotu a unicode hodnotu znaku */
			items = my_strtok(sline_work, 2);
			if (mode == 'U' || mode == 'A') {
				/* pokud je nutne, zvetsi tabulku o TABLE_BUFFER, aby to bylo trochu rychlejsi */
				if (table_length <= table_current_position) {
					table_length = ((table_length/TABLE_BUFFER)+1) * TABLE_BUFFER;
					table = realloc(table, table_length * sizeof(CHARSTRUCT));
					if (table == NULL) error_handler(ERR_ALLOC, "load_table::realloc::1");
				}
				/* urceni modu nacitani: U = unicode kod jako index v tabulce, A = ascii kod jako index v tabulce */
				if (mode == 'U') {
					/* musime osetrit, zda cilovy znak je vubec definovany, napr windows-1250 nema nektrere znaky v unicode kodu */
					if (items[1] != NULL) {
						/* pridame znak do tabulky */
						table[table_current_position].index = strtol(items[1], &chyba, 16);
						table[table_current_position].new_char = strtol(items[0], &chyba, 16);
						table[table_current_position].is_used = 1;
					}
					/* pokud mame nastaveno loadfrom nezvetsime pocitadlo vlozenych znaku, ale primo zpracujeme dalsi znak a aktualni vlozeny prepiseme */
					/* pripadne pokud neni definovany provedeme totez */
					if ((loadfrom != -1 && loadfrom > table[table_current_position].new_char) || items[1] == NULL) {
						memfree(items);
						memfree(sline_small);
						memfree(sline_work);
						continue;
					}
				} else {
					/* musime osetrit, zda cilovy znak je vubec definovany */
					if (items[1] != NULL) {
						table[table_current_position].index = strtol(items[0], &chyba, 16);
						table[table_current_position].new_char = strtol(items[1], &chyba, 16);
						table[table_current_position].is_used = 0;
					}
					/* pokud mame nastaveno loadfrom nezvetsime pocitadlo vlozenych znaku, ale primo zpracujeme dalsi znak a aktualni vlozeny prepiseme */
					/* pripadne pokud neni definovany provedeme totez */
					if ((loadfrom != -1 && loadfrom > table[table_current_position].index) || items[1] == NULL) {
						memfree(items);
						memfree(sline_small);
						memfree(sline_work);
						continue;
					}
				}
				/* provedli jsme uspesny zapsani znaku, zvetsime pocitadlo vlozenych znaku */
				table_current_position++;
			/* mod prirazeni znaku z jineho kodovani k jiz existujicimu znaku v tabulce */
			} else if (mode == 'S') {
				/* musime osetrit, zda cilovy znak je vubec definovany */
				if (items[1] != NULL) {
					/* budeme hledat v table[i].new_char unicode hodnotu t_number */
					t_number = strtol(items[1], &chyba, 16);
					/* osetrime loadfrom */
					if (loadfrom > 0) if (loadfrom > strtol(items[0], &chyba, 16)) {
						memfree(items[0]);
						memfree(items[1]);
						memfree(items);
						memfree(sline_small);
						memfree(sline_work);
						continue;
					}
					/*projedeme celoui tabulku */
					for (i=0;i<table_current_position;i++) {
						/* pokud nalezneme hodnotu table[i].new_char == t_number pak ji prepiseme ascii hodnotou jineho kodovani */
						/* a nastavume priznak, ze se povedlo priradit */
						if (table[i].new_char == t_number && table[i].is_used != 1) {
							table[i].new_char = strtol(items[0], &chyba, 16);
							table[i].is_used = 1;
						}
					}
				}
			}
			/* nakonec provedeme cisteni pameti */
			memfree(items[0]);
			memfree(items[1]);
			memfree(items);
			memfree(sline_small);
			memfree(sline_work);
		}
	}
	/* zmensime tabulku, na takovou velikost, kterou opravdu potrebujeme */
	if (mode == 'U' || mode == 'A') {
		if (table_length != table_current_position) {
			table_length = table_current_position;
			table = realloc(table, table_length * sizeof(CHARSTRUCT));
			if (table == NULL)  error_handler(ERR_ALLOC, "load_table::realloc::2");
		}
	}
	/* uzavreme tabulkovy soubor */
	fclose(fp);
	
}

/**
 * Porovnavaci funkce pro serazeni pole
 */
int cmp_table_index(const void *p_a, const void *p_b) {
	if (((CHARSTRUCT *)p_a)->index < ((CHARSTRUCT *)p_b)->index) return -1;
	else if (((CHARSTRUCT *)p_a)->index == ((CHARSTRUCT *)p_b)->index) return 0;
	else return 1;
}

/**
 * Funkce na serazeni kodovaci tabulky
 */
void sorttable() {
/**
 * Pod linuxem mi to pada na qsortu a vypada to asi na chybu v implementaci, tak 
 * jsem pouzil zdrojaky z gnugo. Jejich implentace by mela byt dokonce
 * rychlejsi, nez originalni qsort na maly pocet dat (aspon to takhle uvadeji).
 */
	if (table_length > 0) gg_sort(table, table_length, sizeof(CHARSTRUCT), cmp_table_index);	
}

/**
 * Funkce na kompresi tabulky - vyhodi hodnoty kterymi nelze provest konverze.
 * Tim se zrychli o neco bsearch.
 */
void compresstable() {
	unsigned int i = 0, x = 0;
	
	for(i = 0; i < table_length; i++) {
		if (table[i].is_used == 1) {
			if (i != x) table[x] = table[i];
			x++;
		}
	}
	table_length = x;
	table = realloc(table, table_length * sizeof(CHARSTRUCT));
}

/**
 * Porovnavaci funkce pro hledani pres bsearch
 */
int sch_index(const void *p_a, const void *p_b) {
	if (*(int *)p_a < ((CHARSTRUCT *)p_b)->index) return -1;
	else if (*(int *)p_a == ((CHARSTRUCT *)p_b)->index) return 0;
	else return 1;
}

/**
 * Funkce pro hledani konverze v poli, pokud je definovano BSEARCH, provede se
 * hledani pres funkci bsearch, jinak se prohleda postupnym prochazeni pole
 */ 
unsigned int search_char(unsigned int index, unsigned int c_default) {
#ifdef BSEARCH
	CHARSTRUCT *p_table_item;
	
	if (table_length == 0) return c_default;
	if (index < table[0].index) return c_default;
	
	if ((p_table_item = (CHARSTRUCT *)bsearch( (void *) &index, (void *)table, table_length, sizeof(CHARSTRUCT), sch_index)) == NULL) return c_default;
	else return p_table_item->new_char;
#else
	int i;
	for (i=0;i<table_length;i++) {
		if (index < table[i].index) return c_default; 
		if (index == table[i].index) return table[i].new_char; 
	}
	return c_default;
#endif
}
