/* -----------------------------------------------------------------------------
 K03-21 Prevadec kodovani ceske diakritiky

 Martin Sloup, A04372
 FAV, INIB-INF
 msloup@students.zcu.cz
 arcao@arcao.com
 Soubor main.c
----------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include "structs.h"
#include "function.h"
#include "file_test.h"
#include "load_table.h"
#include "conv.h"
#include "main.h"

/* Par definic na zacatek */
#define TABLE_PATH "tables\\%s.tab"
#define KONV_VERSION "0.9.9g" 

int param_test_binary; /* Testovat, zda je soubor binarni? 1=Ano, 0=Ne */
int param_newline_char; /* Novy radek? 0=default, 1=crlf, 2=lf, 3=cr (mac) */
int param_utf8_create_bom; /* Vytvorit BOM na zacatku utf-8 souboru? 1=Ano, 0=Ne */
int param_unicode_bom; /* Jak se maji poskladat znaky v UTF-16? 1=BigEndian, 0=LittleEndian */
int param_detect_encoding; /* Bude provedena detekce kodovani? 1=Ano, 0=Ne */

/* FILE struktury pro vstupni a vystupni soubor. */
FILE *fi;
FILE *fo;

/* Promene uchovavajici jmena souboru */
char *filename_table_from;
char *filename_table_to;
char *filename_convert_from;
char *filename_convert_to;

/* Promene pro vlastni vstupni a vystupni bufer */
static char *in_buf;
static char *out_buf;


const char *new_lines[] = {"<CR><LF>", "<LF>", "<CR>"};

void init() {
	/* defaultni hodnoty */

	table_length = 0; /* Velikost [bufferu] prekodovavaci tabulky */
	table_current_position = 0; /* Aktualni pozice v prekodovavaci tabulce */ 
	loadfrom = -1; /* Od jakeho ascii kodu tabulky budemem nacitat? */
	use_internal_encoding_from = 0; /* Pouziti vestavene fce na prekodovani? 0=ascii, 1=utf8, 2=utf16 */
	use_internal_encoding_to = 0; /* Pouziti vestavene fce na prekodovani? 0=ascii, 1=utf8, 2=utf16 */
	
	param_test_binary = 1; /* Testovat, zda je soubor binarni? 1=Ano, 0=Ne */
	param_newline_char = 0; /* Novy radek? 0=default, 1=crlf, 2=lf, 3=cr (mac) */
	param_utf8_create_bom = 0; /* Vytvorit BOM na zacatku utf-8 souboru? 1=Ano, 0=Ne */
	param_detect_encoding = 0; /* Bude provedena detekce kodovani? 1=Ano, 0=Ne */
	param_unicode_bom = 0; /* Jak se maji poskladat znaky v UTF-16? 1=BigEndian, 0=LittleEndian */
}

void showhelp() {
	/* Vyhodi napovedu */
	fprintf(stdout, "Pouziti:\n\n");
	
	fprintf(stdout, "  konv [prepinace] [<vstupni_kodovani>] <vystupni_kodovani> \n");
	fprintf(stdout, "       <vstupni_soubor> <vystupni_soubor> [prepinace]\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "  <vstupni_kodovani> / <vystupni_kodovani> - kodovani definovane v adresari\n");
	fprintf(stdout, "                                             tables\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "  Prepinace:\n");
	fprintf(stdout, "    -b        vypnuti detekce binarniho souboru\n");
	fprintf(stdout, "    -nw       Windowsovske konce radku (<CR><LF>) ve vystupnim souboru\n");
	fprintf(stdout, "    -n[l|u]   Linuxovske konce radku (<LF>) ve vystupnim souboru\n");
	fprintf(stdout, "    -nm       Macovske konce radku (<CR>) ve vystupnim souboru\n");
	fprintf(stdout, "    -u        Vytvori na zacatku souboru v kodovani UTF-8 identifikacni bajty \n");
	fprintf(stdout, "              BOM (Byte Order Mark) - lze pouzit jen na kodovani UTF-8\n");
	fprintf(stdout, "    -obe      Pouzije se Big Endian poradi bytu u UTF-16 (pokud neni\n");
	fprintf(stdout, "              detekovano)\n");
	fprintf(stdout, "    -ole      Pouzije se Little Endian poradi bytu u UTF-16 (pokud neni\n");
	fprintf(stdout, "              detekovano)\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "Pokud neni uvedeno vstupni kodovani, program dokaze rozeznat nasledujici\n");
	fprintf(stdout, "kodovani vstupniho souboru:\n");
	fprintf(stdout, "  iso-8859-2, windows-1250, keybcs2 (kamenicti), koi-8_cs2, latin-2 (DOS\n");
	fprintf(stdout, "  Latin 2), mac-ce (apple-ce), utf-8, utf-16\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "Priklady: \n");
	fprintf(stdout, "  konv utf-8 vstupni.txt vystupni.txt\n");
	fprintf(stdout, "    Provede automatickou detekci vstupniho kodovani a overi, zda je soubor\n");
	fprintf(stdout, "    binarni. Vystupni kodovani bude UTF-8.\n\n");
	fprintf(stdout, "  konv -b utf-8 vstupni.txt vystupni.txt\n");
	fprintf(stdout, "    Provede automatickou detekci vstupniho kodovani, ale neoveri, zda je soubor\n");
	fprintf(stdout, "    binarni. Vystupni kodovani bude UTF-8.\n\n");
	fprintf(stdout, "  konv windows-1250 utf-8 vstupni.txt vystupni.txt\n");
	fprintf(stdout, "    Overi, zda je vstupni soubor binarni a provede konverzi z kodovani\n");
	fprintf(stdout, "    WINDOWS-1250 do UTF-8. Toto je vyhodne, pokud chceme delat konverzi na\n");
	fprintf(stdout, "    neceskych kodovanich.\n\n");
	fprintf(stdout, "  konv -nl utf-8 vstupni.txt vystupni.txt\n");
	fprintf(stdout, "    Overi, zda je vstupni soubor binarni, provede automatickou detekci\n");
	fprintf(stdout, "    vstupniho kodovani. Vystupni bude UTF-8. Pri konverzi pouzije Linuxove\n");
	fprintf(stdout, "    konce radku <LF>.\n\n");
	fprintf(stdout, "  konv -nm -u utf-8 vstupni.txt vystupni.txt\n");
	fprintf(stdout, "    Overi, zda je vstupni soubor binarni, provede automatickou detekci\n");
	fprintf(stdout, "    vstupniho kodovani. Vystupni bude UTF-8. Pri konverzi pouzije Macovske\n");
	fprintf(stdout, "    konce radku <CR> a na zacatek vystupniho souboru vlozi identifikacni bajty\n");
	fprintf(stdout, "    BOM (Byte Order Mark).\n");
}

void parse_param(char *params) {
	/* nechceme tam pomlcku */
	params++;
	
	lcase(params);
	/* osetreni unixoveho --help */
	if (strcmp(params, "-help")==0) params = "h";
		
	switch(toupper(params[0])) {
		case 'B':
			param_test_binary = 0; /* Testovat, zda je soubor binarni? 1=Ano, 0=Ne */
			break;
		case 'N':
			switch(toupper(params[1])) {
				case 'W' :
					param_newline_char = 1; /* Novy radek? 0=default, 1=crlf, 2=lf, 3=cr (mac) */
					break;
				case 'L':
				case 'U':
					param_newline_char = 2; /* Novy radek? 0=default, 1=crlf, 2=lf, 3=cr (mac) */
					break;
				case 'M':
					param_newline_char = 3; /* Novy radek? 0=default, 1=crlf, 2=lf, 3=cr (mac) */
					break;
				default:
					fprintf(stderr, "Chyba: Neznamy prepinac '-%s'!\n", params);
					exit(2);
			}
			break;
		case 'U':
			param_utf8_create_bom = 1; /* Vytvorit BOM na zacatku utf-8 souboru? 1=Ano, 0=Ne */
			break;
		case 'O':
			if (toupper(params[1]) == 'B' && toupper(params[2]) == 'E') param_unicode_bom = 1; /* Jak se maji poskladat znaky v UTF-16? 1=BigEndian, 0=LittleEndian */
			if (toupper(params[1]) == 'L' && toupper(params[2]) == 'E') param_unicode_bom = 0; /* Jak se maji poskladat znaky v UTF-16? 1=BigEndian, 0=LittleEndian */
			break;
		case 'H':
			showhelp();
			exit(0);
			break;
		default:
			fprintf(stderr, "Nastala chyba: Neznamy priznak '-%s'!", params);
			exit(2);
	}
}

void showheader() {
	/* Zobrazi hlavicku programu */
	fprintf(stdout, "konv verze %s\n\n", KONV_VERSION);
	fprintf(stdout, "Autor: Martin Arcao Sloup\n");
	fprintf(stdout, "E-mail: msloup@students.zcu.cz, arcao@arcao.com\n");
	fprintf(stdout, "Tato aplikace vznikla v ramci semestralni prace z predmetu\n");
	fprintf(stdout, "KIV/PT & KIV/PC na Zapadoceske Univerzite v Plzni.\n\n");
}

int main(int argc, char *argv[]) {
	int i;
	char path[PATH_LENGTH];
	
	/* inicializace promenych */
	init();
	
	/* zobrazeni hlavicky */
	showheader();
	
	/* zpracovani vstupnich parametru programu */
	for(i=1;i<argc;i++) {
		if (argv[i][0] == '-') {
			/* zpracovani prepinacu */
			parse_param(argv[i]);
		} else {
			if (filename_table_from == NULL) {
				filename_table_from = trim(argv[i]);
			} else if (filename_table_to == NULL) {
				filename_table_to = trim(argv[i]);
			} else if (filename_convert_from == NULL) {
				filename_convert_from = trim(argv[i]);
			} else if (filename_convert_to == NULL) {
				filename_convert_to = trim(argv[i]);
			}
		}
	}
	/* osetreni ruzneho zpusobu zadani vstupnich parametru - priprava pred autodetekci kodovani */
	if (filename_convert_to == NULL  && filename_table_from != NULL && filename_table_to != NULL && filename_convert_from != NULL) {
		filename_convert_to = filename_convert_from;
		filename_convert_from = filename_table_to;
		filename_table_to = filename_table_from;
		filename_table_from = NULL;
	}
	/* nejsou vsechny potrebne parametry, vypise se napoveda */
	if (filename_table_to == NULL || filename_convert_from == NULL || filename_convert_to == NULL) {
		showhelp();
		return 0;
	}
	/* testovani, zda vstupni soubor je binarni */
	if (param_test_binary == 1) {
		fprintf(stdout, " * Detekuji, zda je soubor binarni... ");
		if (is_file_binary(filename_convert_from) == 1) {
			fprintf(stdout, "BINARNI SOUBOR!\n");
			fprintf(stderr, "CHYBA!!! Vstupni soubor '%s' byl detekovan jako binarni soubor!\n", filename_convert_from);
			fprintf(stderr, "Pro vynechani detekce pouzijte parametr '-b'. Pokud prekodovavate z UTF-16,\n");
			fprintf(stderr, "musite tento parametr pouzit.\n");
			return 1;
		} else {
			fprintf(stdout, "textovy soubor\n");
		}
	}
	/* je-li potreba detekovat vstupni kodovani, detekujeme ho */
	if (filename_table_from == NULL) {
		param_detect_encoding = 1;
		fprintf(stdout, " * Detekuji kodovani vstupniho souboru...");
		fflush(stdout);
		new_string(filename_table_from, 129);
		strcpy(filename_table_from, detect_encoding(filename_convert_from)); 
		fprintf(stdout, "\n * Autodetekce nalezla kodovani %s.\n", filename_table_from);
	}
	/* Nelze provadet konverzi mezi stejnyma kodovanima */
	if (strcmp(filename_table_from, filename_table_to)==0) {
		fprintf(stderr, "CHYBA!!! ZDROJOVE A CILOVE KODOVANI JE TOTOZNE, KONVERZI NELZE PROVEST!\n");
		return 1;
	}
	/* vyhodime par dulezitych zprav */
	fprintf(stdout, " * Nyni provedu konverzi souboru '%s' z kodovani %s do %s.\n", filename_convert_from, filename_table_from, filename_table_to);
	fprintf(stdout, " * Vystup zapisu do souboru '%s' v kodovani %s.\n", filename_convert_to, filename_table_to);
	if (param_newline_char) fprintf(stdout, " * Jako konce radku pouziju %s.\n", new_lines[param_newline_char - 1]);
	
	fprintf(stdout, " * Nacitam kodovaci tabulky...\n");
	
	/* pripravyme jmena souboru tabulek a nacteme tabulky */
	sprintf(path, TABLE_PATH, filename_table_from);
	loadfrom = -1;
	load_table(path, 'a');

	path[0] ='\0';
	sprintf(path, TABLE_PATH, filename_table_to);
	loadfrom = -1;
	load_table(path, 'u');
	
	/* setridime tabulku - nutne, pokud chceme pouzit binarni vyhledavani znaku */
	sorttable();
	
	fprintf(stdout, " * Provadim konverzi...\n");

	/* otevreme soubory */
	if ((fi = fopen(filename_convert_from, "rb")) == NULL) {
		error_handler(ENOENT, "main::fopen::fi");
		fprintf(stderr, "CHYBA!!! Soubor '%s' se nepodarilo otevrit pro zapis.\nSystem hlasi: %s.\n", filename_convert_from, strerror(errno));
		exit(1);
	}
	if ((fo = fopen(filename_convert_to, "wb")) == NULL) {
		error_handler(ENOENT, "main::fopen::fo");
		fprintf(stderr, "CHYBA!!! Soubor '%s' se nepodarilo otevrit pro zapis.\nSystem hlasi: %s.\n", filename_convert_to, strerror(errno));
		exit(1);
	}
	/* Nastavyme vstupni / vystupni bufer */
	fsetbuf(fi, in_buf, 16 * 1024);
	fsetbuf(fo, out_buf, 16 * 1024);

	/* provedem konverzi - zaklezi na typu kodovani */
	if (use_internal_encoding_from == 0 && use_internal_encoding_to == 0) {
		/* provedem kompresi tabulky - vyhodime znaky na ktere nelze provest konverze */
		compresstable();
		/* konverze ascii to ascii - zalezi na nactene tabulce */
		conv_ascii2ascii();
	} else if (use_internal_encoding_from == 0 && use_internal_encoding_to == 1) {
		/* konverze ascii to utf-8 - pro ascii je nutna tabulka znaku */
		conv_ascii2utf8();
	} else if (use_internal_encoding_from == 1 && use_internal_encoding_to == 0) {
		/* konverze utf-8 to ascii - pro ascii je nutna tabulka znaku */
		conv_utf82ascii();
	} else if (use_internal_encoding_from == 0 && use_internal_encoding_to == 2) {
		/* konverze ascii to unicode - pro ascii je nutna tabulka znaku */
		conv_ascii2unicode();
	} else if (use_internal_encoding_from == 2 && use_internal_encoding_to == 0) {
		/* konverze unicode to ascii - pro ascii je nutna tabulka znaku */
		conv_unicode2ascii();
	} else if (use_internal_encoding_from == 1 && use_internal_encoding_to == 2) {
		/* konverze utf-8 to unicode - dekodujem znak z utf-8 a ulozime ho primo do souboru */
		conv_utf82unicode();
	} else if (use_internal_encoding_from == 2 && use_internal_encoding_to == 1) {
		/* konverze unicode to utf-8 - nactem znak ze souboru a ulozime ho zakodovany v utf-8 */
		conv_unicode2utf8();
	} else {
		fprintf(stderr, "Omlovam se, tuto konverzi prozatim neumim.\n");
	}

	/* uzavreme soubory a uvolnime pouzitou pamet */
	fclose(fi);
	fclose(fo);
	memfree(in_buf);
	memfree(out_buf);
	
	memfree(table);
	table_length = 0;
	memfree(filename_table_from);
	memfree(filename_table_to);
	memfree(filename_convert_from);
	memfree(filename_convert_to);

	return 0;
}
